import 'dart:async';
import 'dart:io';
import 'package:path/path.dart';
import 'package:path_provider/path_provider.dart';
import 'package:sembast/sembast_io.dart';
import 'package:natinfo_flutter/api/swagger.swagger.dart';

class DatabaseHelper {
  static final DatabaseHelper _singleton = DatabaseHelper._internal();
  factory DatabaseHelper() => _singleton;
  DatabaseHelper._internal();

  Database? _db;
  final _store = intMapStoreFactory.store('natinf');
  final _historyStore = intMapStoreFactory.store('historique');

  /// Initialise la base de données Sembast
  Future<void> initDb() async {
    final Directory dir = await getApplicationDocumentsDirectory();
    final dbPath = join(dir.path, 'natinfo.db');
    _db = await databaseFactoryIo.openDatabase(dbPath);
  }

  Future<int> insertOrUpdateNatinf(Natinf natinf) async {
    final record = _store.record(natinf.id!);
    await record.put(_db!, natinf.toJson());
    return natinf.id!;
  }

  Future<Natinf?> getNatinfByNumero(String numero) async {
    if (_db == null) throw Exception("La base de données n'est pas initialisée.");
    final finder = Finder(filter: Filter.equals('numero_natinf', numero));
    final record = await _store.findFirst(_db!, finder: finder);
    if (record != null) {
      return Natinf.fromJson(record.value);
    }
    return null;
  }

  /// Recherche tous les natinf dont l’un des champs (numero_natinf,
  /// nature_infraction, qualification_infraction) contient la requête.
  /// La recherche est permissive (ignore majuscules/minuscules, ponctuation, etc.).
  Future<List<Natinf>> getNatinfByQuery(String query) async {
    // Fonction de normalisation : convertir en minuscules et remplacer les caractères nonR alphanumériques par un espace.
    String normalize(String s) => s.toLowerCase().replaceAll(RegExp(r'[^\w]+'), ' ').trim();
    
    final normalizedQuery = normalize(query);
    // Sépare la requête normalisée en mots
    final queryWords = normalizedQuery.split(' ').where((word) => word.isNotEmpty).toList();
    
    // Utilisation d'un filtre personnalisé pour tester si tous les mots de la requête apparaissent dans au moins un des champs combinés.
    final finder = Finder(filter: Filter.custom((record) {
      final Map<String, dynamic> value = record.value as Map<String, dynamic>;
      final numero = normalize(value['numero_natinf']?.toString() ?? '');
      final nature = normalize(value['nature_infraction']?.toString() ?? '');
      final qualification = normalize(value['qualification_infraction']?.toString() ?? '');
      final definiePar = normalize(value['definie_par']?.toString() ?? '');
      final reprimeePar = normalize(value['reprimee_par']?.toString() ?? '');
      // Concatène les valeurs de plusieurs champs pour chercher tous les mots
      final combined = '$numero $nature $qualification $definiePar $reprimeePar';
      // Vérifie que chaque mot de la requête se trouve dans les champs combinés
      return queryWords.every((word) => combined.contains(word));
    }));
    
    final records = await _store.find(_db!, finder: finder);
    return records.map((record) => Natinf.fromJson(record.value)).toList();
  }

  Future<int> addHistory(String numero) async {
    if (_db == null) throw Exception("La base de données n'est pas initialisée.");
    return await _historyStore.add(_db!, {
      'numero_natinf': numero,
      'opened_at': DateTime.now().millisecondsSinceEpoch,
    });
  }

  Future<List<Map<String, dynamic>>> getHistory() async {
    if (_db == null) throw Exception("La base de données n'est pas initialisée.");
    final finder = Finder(sortOrders: [SortOrder('opened_at', false)]);
    final records = await _historyStore.find(_db!, finder: finder);
    return records.map((record) => record.value as Map<String, dynamic>).toList();
  }

  Future<void> clearHistory() async {
    if (_db == null) throw Exception("La base de données n'est pas initialisée.");
    await _historyStore.delete(_db!);
  }

  Future<void> clearOfflineData() async {
    if (_db == null) throw Exception("La base de données n'est pas initialisée.");
    await _store.delete(_db!);
    print("Données locales supprimées.");
  }

  Future<bool> isEmpty() async {
    if (_db == null) {
      throw Exception("La base de données n'est pas initialisée.");
    }
    final count = await _store.count(_db!);
    return count == 0;
  }
}