import 'package:flutter/material.dart';
import 'package:natinfo_flutter/api/swagger.swagger.dart';
import 'package:natinfo_flutter/db/database_helper.dart';
import 'package:natinfo_flutter/widgets/natinf_list.dart';

enum SortOption {
  defaultOrder,
  oldestFirst,
  mostConsulted,
  leastConsulted,
}

class HistoriquePage extends StatefulWidget {
  const HistoriquePage({super.key});

  @override
  _HistoriquePageState createState() => _HistoriquePageState();
}

class _HistoriquePageState extends State<HistoriquePage> {
  List<Map<String, dynamic>> _historyRecords = [];
  List<Natinf> _natinfHistory = [];
  SortOption _sortOption = SortOption.defaultOrder;
  bool _isLoading = true;

  @override
  void initState() {
    super.initState();
    _loadHistory();
  }

  Future<void> _loadHistory() async {
    final records = await DatabaseHelper().getHistory();
    final futures = records.map((rec) async {
      return await DatabaseHelper().getNatinfByNumero(rec['numero_natinf']);
    }).toList();
    final results = await Future.wait(futures);
    setState(() {
      _historyRecords = records;
      _natinfHistory = results.whereType<Natinf>().toList();
      _isLoading = false;
    });
  }

  Future<void> _confirmClearHistory() async {
    final shouldClear = await showDialog<bool>(
      context: context,
      builder: (context) => AlertDialog(
        title: const Text("Confirmer la suppression"),
        content: const Text("Voulez-vous vraiment supprimer l'historique ?"),
        actions: [
          TextButton(
            onPressed: () => Navigator.pop(context, false),
            child: const Text("Annuler"),
          ),
          TextButton(
            onPressed: () => Navigator.pop(context, true),
            child: const Text("Supprimer"),
          ),
        ],
      ),
    );
    if (shouldClear == true) {
      await DatabaseHelper().clearHistory();
      setState(() {
        _historyRecords.clear();
        _natinfHistory.clear();
      });
    }
  }

  List<Natinf> _applySort() {
    switch (_sortOption) {
      case SortOption.oldestFirst:
        return List.from(_natinfHistory.reversed);

      case SortOption.mostConsulted:
      case SortOption.leastConsulted:
        final counts = <String, int>{};
        for (var rec in _historyRecords) {
          final key = rec['numero_natinf'].toString();
          counts[key] = (counts[key] ?? 0) + 1;
        }
        final mapN = <String, Natinf>{};
        for (var n in _natinfHistory) {
          final key = n.numeroNatinf?.toString() ?? '';
          mapN[key] = n;
        }
        final entries = counts.entries.toList();
        entries.sort((a, b) {
          if (_sortOption == SortOption.mostConsulted) {
            return b.value.compareTo(a.value);
          } else {
            return a.value.compareTo(b.value);
          }
        });
        return entries.map((e) => mapN[e.key]!).toList();

      case SortOption.defaultOrder:
      default:
        return List.from(_natinfHistory);
    }
  }

  @override
  Widget build(BuildContext context) {
    final displayList = _applySort();
    return Scaffold(
      appBar: AppBar(
        title: const Text("Historique"),
        actions: [
          IconButton(
            icon: const Icon(Icons.delete),
            onPressed: _confirmClearHistory,
          )
        ],
      ),
      body: _isLoading
          ? const Center(child: CircularProgressIndicator())
          : Padding(
              padding: const EdgeInsets.all(16),
              child: Column(
                crossAxisAlignment: CrossAxisAlignment.start,
                children: [
                  DropdownButtonFormField<SortOption>(
                    value: _sortOption,
                    decoration: InputDecoration(
                      labelText: 'Trier par',
                      border: OutlineInputBorder(borderRadius: BorderRadius.circular(8)),
                      contentPadding: EdgeInsets.symmetric(horizontal: 12, vertical: 4),
                    ),
                    onChanged: (val) {
                      if (val != null) setState(() => _sortOption = val);
                    },
                    items: [
                      DropdownMenuItem(
                        value: SortOption.defaultOrder,
                        child: Text('Du plus récent au plus ancien'),
                      ),
                      DropdownMenuItem(
                        value: SortOption.oldestFirst,
                        child: Text('Du plus ancien au plus récent'),
                      ),
                      DropdownMenuItem(
                        value: SortOption.mostConsulted,
                        child: Text('Les plus consultés'),
                      ),
                      DropdownMenuItem(
                        value: SortOption.leastConsulted,
                        child: Text('Les moins consultés'),
                      ),
                    ],
                  ),
                  const SizedBox(height: 8),
                  Expanded(
                    child: Container(
                      margin: const EdgeInsets.only(top: 8.0),
                      decoration: BoxDecoration(
                        borderRadius: BorderRadius.circular(8),
                      ),
                      child: ClipRRect(
                        borderRadius: BorderRadius.circular(8),
                        child: NatinfList(
                          natinfList: displayList,
                          recordHistory: false,
                        ),
                      ),
                    ),
                  ),
                ],
              ),
            ),
    );
  }
}
